@@ -21,8 +21,10 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <dirent.h>
 #include <sys/stat.h>
 #include <sys/unistd.h>
+#include <sys/types.h>
 #include <unistd.h>
 
 #define	FLAGS_TEST	(1 << 0)
@@ -55,9 +57,9 @@ struct alternativeSet {
     char * currentLink;
 };
 
-enum programModes { MODE_UNKNOWN, MODE_INSTALL, MODE_REMOVE, MODE_AUTO, 
-		    MODE_DISPLAY, MODE_CONFIG, MODE_SET, 
-		    MODE_SLAVE, MODE_VERSION, MODE_USAGE };
+enum programModes { MODE_UNKNOWN, MODE_INSTALL, MODE_REMOVE, MODE_AUTO,
+		    MODE_DISPLAY, MODE_CONFIG, MODE_SET,
+		    MODE_SLAVE, MODE_VERSION, MODE_USAGE, MODE_LIST };
 
 static int usage(int rc) {
     printf(_("alternatives version %s - Copyright (C) 2001 Red Hat, Inc.\n"), VERSION);
@@ -70,6 +72,7 @@ static int usage(int rc) {
     printf(_("       alternatives --config <name>\n"));
     printf(_("       alternatives --display <name>\n"));
     printf(_("       alternatives --set <name> <path>\n"));
+    printf(_("       alternatives --list\n"));
     printf(_("\n"));
     printf(_("common options: --verbose --test --help --usage --version\n"));
     printf(_("                --altdir <directory> --admindir <directory>\n"));
@@ -77,10 +80,25 @@ static int usage(int rc) {
     exit(rc);
 }
 
-static void setupSingleArg(enum programModes * mode, const char *** nextArgPtr, 
+static int isSystemd(char *initscript) {
+    char tmppath[500];
+    struct stat sbuf;
+
+    snprintf(tmppath, 500, "/lib/systemd/system/%s.service", initscript);
+    if (!stat(tmppath, &sbuf))
+        return 1;
+
+    snprintf(tmppath, 500, "/etc/systemd/system/%s.service", initscript);
+    if (!stat(tmppath, &sbuf))
+        return 1;
+
+    return 0;
+}
+
+static void setupSingleArg(enum programModes * mode, const char *** nextArgPtr,
 			   enum programModes newMode, char ** title) {
     const char ** nextArg = *nextArgPtr;
-    
+
     if (*mode != MODE_UNKNOWN) usage(2);
     *mode = newMode;
     nextArg++;
@@ -90,7 +108,7 @@ static void setupSingleArg(enum programModes * mode, const char *** nextArgPtr,
     *nextArgPtr = nextArg + 1;
 }
 
-static void setupDoubleArg(enum programModes * mode, const char *** nextArgPtr, 
+static void setupDoubleArg(enum programModes * mode, const char *** nextArgPtr,
 			   enum programModes newMode, char ** title,
 			   char ** target) {
     const char ** nextArg = *nextArgPtr;
@@ -143,7 +161,7 @@ char * parseLine(char ** buf) {
     return strdup(start);
 }
 
-static int readConfig(struct alternativeSet * set, const char * title, 
+static int readConfig(struct alternativeSet * set, const char * title,
 		      const char * altDir, const char * stateDir, int flags) {
     char * path;
     int fd;
@@ -255,10 +273,10 @@ static int readConfig(struct alternativeSet * set, const char * title,
 	set->alts[set->numAlts].master.facility = strdup(groups[0].facility);
 	set->alts[set->numAlts].master.title = strdup(groups[0].title);
 	set->alts[set->numAlts].master.target = line;
-	set->alts[set->numAlts].numSlaves = numGroups - 1; 
+	set->alts[set->numAlts].numSlaves = numGroups - 1;
 	if (numGroups > 1)
-	    set->alts[set->numAlts].slaves = 
-		malloc((numGroups - 1) * 
+	    set->alts[set->numAlts].slaves =
+		malloc((numGroups - 1) *
 			sizeof(*set->alts[set->numAlts].slaves));
 	else
 	    set->alts[set->numAlts].slaves = NULL;
@@ -273,15 +291,9 @@ static int readConfig(struct alternativeSet * set, const char * title,
 	    return 1;
 	}
 	if (end) {
-		char tmppath[500];
-		struct stat sbuf;
-		
 		while (*end && isspace(*end)) end++;
 		if (strlen(end)) {
-			snprintf(tmppath, 500, "/etc/init.d/%s", end);
-			if (!stat(tmppath, &sbuf)) {
-				set->alts[set->numAlts].initscript = strdup(end);
-			}
+                        set->alts[set->numAlts].initscript = strdup(end);
 		}
 	}
 
@@ -296,9 +308,9 @@ static int readConfig(struct alternativeSet * set, const char * title,
 		return 1;
 	    }
 
-	    set->alts[set->numAlts].slaves[i - 1].title = 
+	    set->alts[set->numAlts].slaves[i - 1].title =
 			strdup(groups[i].title);
-	    set->alts[set->numAlts].slaves[i - 1].facility = 
+	    set->alts[set->numAlts].slaves[i - 1].facility =
 			strdup(groups[i].facility);
 		set->alts[set->numAlts].slaves[i - 1].target = (line && strlen(line)) ? line : NULL;
 	}
@@ -351,7 +363,7 @@ static int readConfig(struct alternativeSet * set, const char * title,
 static int isLink(char *path)  {
 	struct stat sbuf;
 	int rc = 0;
-	
+
 	rc = lstat(path, &sbuf);
 	if (!rc) {
 		rc = S_ISLNK(sbuf.st_mode);
@@ -379,13 +391,12 @@ static int removeLinks(struct linkSet * l, const char * altDir, int flags) {
 		l->facility, strerror(errno));
 	return 1;
     }
-    
+
     return 0;
 }
 
 static int makeLinks(struct linkSet * l, const char * altDir, int flags) {
     char * sl;
-    struct stat sb;
 
     sl = alloca(strlen(altDir) + strlen(l->title) + 2);
     sprintf(sl, "%s/%s", altDir, l->title);
@@ -401,8 +412,10 @@ static int makeLinks(struct linkSet * l, const char * altDir, int flags) {
 			    return 1;
 		    }
 	    }
-    }
-	
+    } else
+            fprintf(stderr, _("failed to link %s -> %s: %s exists and it is not a symlink\n"),
+				    l->facility, sl, l->facility);
+
     if (FL_TEST(flags)) {
 	printf(_("would link %s -> %s\n"), sl, l->target);
     } else {
@@ -410,15 +423,15 @@ static int makeLinks(struct linkSet * l, const char * altDir, int flags) {
 	    fprintf(stderr, _("failed to remove link %s: %s\n"),
 		    sl, strerror(errno));
 	    return 1;
-	} 
-	
+	}
+
 	if (symlink(l->target, sl)) {
 	    fprintf(stderr, _("failed to link %s -> %s: %s\n"),
 		    sl, l->target, strerror(errno));
 	    return 1;
 	}
     }
-    
+
     return 0;
 }
 
@@ -431,7 +444,7 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
     int i, j;
     int rc = 0;
     struct alternative * alt;
-    
+
     path = alloca(strlen(stateDir) + strlen(set->alts[0].master.title) + 6);
     sprintf(path, "%s/%s.new", stateDir, set->alts[0].master.title);
 
@@ -444,10 +457,10 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
 	fd = open(path, O_RDWR | O_CREAT | O_EXCL, 0644);
 
     if (fd < 0) {
-	if (errno == EEXIST) 
+	if (errno == EEXIST)
 	    fprintf(stderr, _("%s already exists\n"), path);
-	else 
-	    fprintf(stderr, _("failed to create %s: %s\n"), path, 
+	else
+	    fprintf(stderr, _("failed to create %s: %s\n"), path,
 		    strerror(errno));
 	return 1;
     }
@@ -474,7 +487,7 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
 		fprintf(f,"\n");
 	}
     }
-    
+
     fclose(f);
 
     if (!FL_TEST(flags) && rename(path, path2)) {
@@ -486,7 +499,7 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
 
     if (set->mode == AUTO)
 		set->current = set->best;
-    
+
     alt = set->alts + ( set->current > 0 ? set->current : 0);
 
     if (forceLinks || set->mode == AUTO) {
@@ -501,20 +514,36 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
 
     if (!FL_TEST(flags)) {
 	    if (alt->initscript) {
-		    path = alloca(strlen("/sbin/chkconfig --add ") + strlen(alt->initscript) + 1);
-		    sprintf(path, "/sbin/chkconfig --add %s", alt->initscript);
-		    if (FL_VERBOSE(flags))
-			    printf(_("running %s\n"), path);
-		    system(path);
+	            if (isSystemd(alt->initscript)) {
+                            asprintf(&path, "/bin/systemctl -q is-enabled %s.service || /bin/systemctl -q preset %s.service", alt->initscript, alt->initscript);
+                            if (FL_VERBOSE(flags))
+                                    printf(_("running %s\n"), path);
+                            system(path);
+                            free(path);
+                    } else {
+                            asprintf(&path, "/sbin/chkconfig --add %s", alt->initscript);
+                            if (FL_VERBOSE(flags))
+                                    printf(_("running %s\n"), path);
+                            system(path);
+                            free(path);
+                    }
 	    }
 	    for (i = 0; i < set->numAlts ; i++) {
 		    struct alternative * tmpalt = set->alts + i;
 		    if (tmpalt != alt && tmpalt->initscript) {
-			    path = alloca(strlen("/sbin/chkconfig --del ") + strlen(tmpalt->initscript) + 1);
-			    sprintf(path, "/sbin/chkconfig --del %s", tmpalt->initscript);
-			    if (FL_VERBOSE(flags))
-				    printf(_("running %s\n"), path);
-			    system(path);
+		            if (isSystemd(tmpalt->initscript)) {
+                                    asprintf(&path, "/bin/systemctl -q disable %s.service", tmpalt->initscript);
+                                    if (FL_VERBOSE(flags))
+				            printf(_("running %s\n"), path);
+                                    system(path);
+                                    free(path);
+                            } else {
+                                    asprintf(&path, "/sbin/chkconfig --del %s", tmpalt->initscript);
+                                    if (FL_VERBOSE(flags))
+				            printf(_("running %s\n"), path);
+                                    system(path);
+                                    free(path);
+                            }
 		    }
 	    }
     }
@@ -524,7 +553,7 @@ static int writeState(struct alternativeSet *  set, const char * altDir,
 
 static int linkCmp(const void *a, const void *b) {
 	struct linkSet *one = (struct linkSet *)a, *two = (struct linkSet *)b;
-	
+
 	return strcmp(one->facility, two->facility);
 }
 
@@ -535,7 +564,7 @@ static int addService(struct alternative newAlt, const char * altDir,
     struct linkSet * newLinks;
     int i, j, k, rc;
 
-    if ( (rc=readConfig(&set, newAlt.master.title, altDir, stateDir, flags)) && rc != 3 && rc != 2) 
+    if ( (rc=readConfig(&set, newAlt.master.title, altDir, stateDir, flags)) && rc != 3 && rc != 2)
 	return 2;
 
     if (set.numAlts) {
@@ -544,11 +573,11 @@ static int addService(struct alternative newAlt, const char * altDir,
 		    set.alts[0].master.title, set.alts[0].master.facility);
 	    return 2;
 	}
-	
+
         /* Determine the maximal set of slave links. */
         base.numSlaves = set.alts[0].numSlaves;
 	base.slaves = malloc(set.alts[0].numSlaves * sizeof(struct linkSet));
-	memcpy(base.slaves, set.alts[0].slaves, 
+	memcpy(base.slaves, set.alts[0].slaves,
 	       set.alts[0].numSlaves * sizeof(struct linkSet));
 
 	for (i = 0; i < newAlt.numSlaves; i++) {
@@ -564,14 +593,14 @@ static int addService(struct alternative newAlt, const char * altDir,
 			base.numSlaves++;
 		}
 	}
-	
+
 	/* Insert new set into the set of alternatives */
 	for (i = 0 ; i < set.numAlts ; i++) {
 		if (!strcmp(set.alts[i].master.target, newAlt.master.target)) {
 			set.alts[i] = newAlt;
 			break;
 		}
-		
+
 	}
 	if (i == set.numAlts) {
 		set.alts = realloc(set.alts, sizeof(*set.alts) * (set.numAlts + 1));
@@ -590,7 +619,7 @@ static int addService(struct alternative newAlt, const char * altDir,
 					    && set.alts[j].slaves[k].target)
 						goto found;
 				}
-				
+
 			}
 			removeLinks(base.slaves + i, altDir, flags);
 			base.numSlaves--;
@@ -602,16 +631,16 @@ static int addService(struct alternative newAlt, const char * altDir,
 			i++;
 		}
 	}
-	
+
 	/* Sort the list for file legibility */
 	qsort(base.slaves, base.numSlaves, sizeof(struct linkSet), linkCmp);
-	    
+
 	/* need to match the slaves up; newLinks will parallel the original
 	   ordering */
 	for (k = 0; k < set.numAlts ; k++) {
 		newLinks = alloca(sizeof(*newLinks) * base.numSlaves);
 		newLinks = memset(newLinks, 0, sizeof(*newLinks) * base.numSlaves);
-	        
+
 		for (j = 0; j < base.numSlaves; j++) {
 			for (i = 0; i < set.alts[k].numSlaves; i++) {
 				if (!strcmp(set.alts[k].slaves[i].title,
@@ -645,7 +674,7 @@ static int addService(struct alternative newAlt, const char * altDir,
 		    set.best = set.numAlts;
 	    set.numAlts++;
     }
-    
+
     if (writeState(&set, altDir, stateDir, 0, flags)) return 2;
 
     return 0;
@@ -654,7 +683,7 @@ static int addService(struct alternative newAlt, const char * altDir,
 static int displayService(char * title, const char * altDir,
 		          const char * stateDir, int flags) {
     struct alternativeSet set;
-	
+
     int alt;
     int slave;
 
@@ -662,7 +691,7 @@ static int displayService(char * title, const char * altDir,
 
     if (set.mode == AUTO)
 	printf(_("%s - status is auto.\n"), title);
-    else	
+    else
 	printf(_("%s - status is manual.\n"), title);
 
     printf(_(" link currently points to %s\n"), set.currentLink);
@@ -682,7 +711,7 @@ static int displayService(char * title, const char * altDir,
     return 0;
 }
 
-static int autoService(char * title, const char * altDir, 
+static int autoService(char * title, const char * altDir,
 		       const char * stateDir, int flags) {
     struct alternativeSet set;
 
@@ -695,7 +724,7 @@ static int autoService(char * title, const char * altDir,
     return 0;
 }
 
-static int configService(char * title, const char * altDir, 
+static int configService(char * title, const char * altDir,
 		       const char * stateDir, int flags) {
     struct alternativeSet set;
     int i;
@@ -746,8 +775,8 @@ static int setService(const char * title, const char * target,
 	if (!strcmp(set.alts[i].master.target, target)) break;
 
     if (i == set.numAlts) {
-	fprintf(stderr, 
-	        _("%s has not been configured as an alternative for %s\n"), 
+	fprintf(stderr,
+	        _("%s has not been configured as an alternative for %s\n"),
 		target, title);
 	return 2;
     }
@@ -772,8 +801,8 @@ static int removeService(const char * title, const char * target,
 	if (!strcmp(set.alts[i].master.target, target)) break;
 
     if (i == set.numAlts) {
-	fprintf(stderr, 
-	        _("%s has not been configured as an alternative for %s\n"), 
+	fprintf(stderr,
+	        _("%s has not been configured as an alternative for %s\n"),
 		target, title);
 	return 2;
     }
@@ -791,11 +820,11 @@ static int removeService(const char * title, const char * target,
 	if (FL_TEST(flags)) {
 	    printf(_("(would remove %s\n"), path);
 	} else if (unlink(path)) {
-	    fprintf(stderr, _("failed to remove %s: %s\n"), path, 
+	    fprintf(stderr, _("failed to remove %s: %s\n"), path,
 		    strerror(errno));
 	    rc |= 1;
 	}
-	
+
 	if (rc) return 2; else return 0;
     }
 
@@ -824,6 +853,31 @@ static int removeService(const char * title, const char * target,
     return 0;
 }
 
+static int listServices(const char * altDir, const char * stateDir, int flags) {
+        DIR *dir;
+        struct dirent *ent;
+        dir = opendir(stateDir);
+        struct alternativeSet set;
+
+        if(dir == NULL)
+                return 2;
+
+        while((ent = readdir(dir)) != NULL) {
+                if(!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
+                        continue;
+
+                if (readConfig(&set, ent->d_name, altDir, stateDir, flags))
+                        return 2;
+
+                printf("%s\t%s\t%s\n", ent->d_name, set.mode == AUTO?"auto":"manual", set.currentLink);
+        }
+
+        closedir(dir);
+
+        return 0;
+}
+
+
 int main(int argc, const char ** argv) {
     const char ** nextArg;
     char * end;
@@ -835,9 +889,9 @@ int main(int argc, const char ** argv) {
     char * stateDir = "/var/lib/alternatives";
     struct stat sb;
 
-    setlocale(LC_ALL, ""); 
-    bindtextdomain("chkconfig","/usr/share/locale"); 
-    textdomain("chkconfig"); 
+    setlocale(LC_ALL, "");
+    bindtextdomain("chkconfig","/usr/share/locale");
+    textdomain("chkconfig");
 
     if (!argv[1])
 	return usage(2);
@@ -867,7 +921,7 @@ int main(int argc, const char ** argv) {
 	} else if (!strcmp(*nextArg, "--initscript")) {
 	    if (mode != MODE_UNKNOWN && mode != MODE_INSTALL) usage(2);
 	    nextArg++;
-	    
+
 	    if (!*nextArg) usage(2);
 	    newAlt.initscript = strdup(*nextArg);
 	    nextArg++;
@@ -881,12 +935,12 @@ int main(int argc, const char ** argv) {
 	    setupSingleArg(&mode, &nextArg, MODE_DISPLAY, &title);
 	} else if (!strcmp(*nextArg, "--config")) {
 	    setupSingleArg(&mode, &nextArg, MODE_CONFIG, &title);
-	} else if (!strcmp(*nextArg, "--help") || 
+	} else if (!strcmp(*nextArg, "--help") ||
 		   !strcmp(*nextArg, "--usage")) {
 	    if (mode != MODE_UNKNOWN) usage(2);
 	    mode = MODE_USAGE;
 	    nextArg++;
-	} else if (!strcmp(*nextArg, "--test")) { 
+	} else if (!strcmp(*nextArg, "--test")) {
 	    flags |= FLAGS_TEST;
 	    nextArg++;
 	} else if (!strcmp(*nextArg, "--verbose")) {
@@ -906,6 +960,10 @@ int main(int argc, const char ** argv) {
 	    if (!*nextArg) usage(2);
 	    stateDir = strdup(*nextArg);
 	    nextArg++;
+        } else if (!strcmp(*nextArg, "--list")) {
+	    if (mode != MODE_UNKNOWN) usage(2);
+	    mode = MODE_LIST;
+	    nextArg++;
 	} else {
 	    usage(2);
 	}
@@ -945,6 +1003,8 @@ int main(int argc, const char ** argv) {
 	return removeService(title, target, altDir, stateDir, flags);
       case MODE_SLAVE:
 	usage(2);
+      case MODE_LIST:
+        return listServices(altDir, stateDir, flags);
     }
 
     abort();
