@@ -19,25 +19,58 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <glob.h>
-#include <libintl.h> 
-#include <locale.h> 
+#include <libintl.h>
+#include <locale.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
+#include <selinux/selinux.h>
+#include <selinux/label.h>
+#include <libgen.h>
 
 /* Changes
    1998-09-22 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
                 i18n for init.d scripts (eg.: description(pt_BR) is a brazilian
-		portuguese description for the package) 
+		portuguese description for the package)
 */
 
-#define _(String) gettext((String)) 
+#define _(String) gettext((String))
 
 #include "leveldb.h"
 
+int selinux_restore(const char *name) {
+        struct selabel_handle *hnd = NULL;
+        struct stat buf;
+        security_context_t newcon = NULL;
+        int r = -1;
+
+        hnd = selabel_open(SELABEL_CTX_FILE, NULL, 0);
+        if (hnd == NULL)
+                goto out;
+
+        r = stat(name, &buf);
+        if (r < 0)
+                goto out;
+
+        r = selabel_lookup_raw(hnd, &newcon, name, buf.st_mode);
+        if (r < 0)
+                goto out;
+
+        r = setfilecon_raw(name, newcon);
+        if (r < 0)
+                goto out;
+
+        r = 0;
+
+ out:
+        selabel_close(hnd);
+        freecon(newcon);
+        return r;
+}
+
 int parseLevels(char * str, int emptyOk) {
     char * chptr = str;
     int rc = 0;
@@ -54,7 +87,7 @@ int parseLevels(char * str, int emptyOk) {
     return rc;
 }
 
-int readDescription(char *start, char *bufstop, char **english_desc, char **serv_desc) { 
+int readDescription(char *start, char *bufstop, char **english_desc, char **serv_desc) {
 	char english;
 	char my_lang_loaded = 0;
 	char is_my_lang = 0;
@@ -62,10 +95,10 @@ int readDescription(char *start, char *bufstop, char **english_desc, char **serv
 	char * final_parenthesis;
 	char * end, *next;
 	int i;
-	
+
 	english = *start == ':';
 	end = strchr(start, '\n');
-	if (!end) 
+	if (!end)
 	    next = end = bufstop;
 	else
 	    next = end + 1;
@@ -100,7 +133,7 @@ int readDescription(char *start, char *bufstop, char **english_desc, char **serv
 	    while (desc[strlen(desc) - 1] == '\\') {
 		desc[strlen(desc) - 1] = '\0';
 		start = next;
-		
+
 		while (isspace(*start) && start < bufstop) start++;
 		if (start == bufstop || *start != '#') {
 		    return 1;
@@ -114,7 +147,7 @@ int readDescription(char *start, char *bufstop, char **english_desc, char **serv
 		}
 
 		end = strchr(start, '\n');
-		if (!end) 
+		if (!end)
 		    next = end = bufstop;
 		else
 		    next = end + 1;
@@ -150,39 +183,35 @@ int readDescription(char *start, char *bufstop, char **english_desc, char **serv
 	return 0;
 }
 
-int readXinetdServiceInfo(char *name, struct service * service, int honorHide) {
-	char * filename = alloca(strlen(name) + strlen(XINETDDIR) + 50);
+int readXinetdServiceInfo(char *name, struct service * service) {
+	char * filename;
 	int fd;
-	struct service serv = { 
+	struct service serv = {
 			name: NULL,
 			levels: -1,
-			kPriority: -1,
-			sPriority: -1, 
-			desc: NULL, 
-			startDeps: NULL, 
+			kPriority: 100,
+			sPriority: -1,
+			desc: NULL,
+			startDeps: NULL,
 			stopDeps: NULL,
 			softStartDeps: NULL,
 			softStopDeps: NULL,
 		        provides: NULL,
-			type: TYPE_XINETD, 
-			isLSB: 0, 
+			type: TYPE_XINETD,
+			isLSB: 0,
 			enabled: -1
 	};
 	struct stat sb;
-	char * buf, *ptr;
+	char * buf = NULL, *ptr;
 	char * eng_desc = NULL, *start;
-	
-	snprintf(filename, strlen(name)+strlen(XINETDDIR)+50, XINETDDIR "/%s", name);
-	
-	if ((fd = open(filename, O_RDONLY)) < 0) return -1;
+
+	asprintf(&filename, XINETDDIR "/%s", name);
+
+	if ((fd = open(filename, O_RDONLY)) < 0) goto out_err;
 	fstat(fd,&sb);
-	if (! S_ISREG(sb.st_mode)) return -1;
+	if (! S_ISREG(sb.st_mode)) goto out_err;
 	buf = malloc(sb.st_size+1);
-	if (read(fd,buf,sb.st_size)!=sb.st_size) {
-		close(fd);
-		free(buf);
-		return -1;
-	}
+	if (read(fd,buf,sb.st_size)!=sb.st_size) goto out_err;
 	close(fd);
         serv.name = strdup(name);
 	buf[sb.st_size] = '\0';
@@ -217,14 +246,14 @@ int readXinetdServiceInfo(char *name, struct service * service, int honorHide) {
 			if (ptr) {
 				*ptr = '\0';
 				ptr++;
-			} 
+			}
 			buf = ptr;
 			continue;
 		}
 		while (isspace(*buf) && buf < ptr) buf++;
 		if (!strncmp(buf,"disable", 7)) {
 			buf = strstr(buf,"=");
-			if (buf) 
+			if (buf)
 			  do {
 				  buf++;
 			  } while(isspace(*buf));
@@ -242,11 +271,17 @@ int readXinetdServiceInfo(char *name, struct service * service, int honorHide) {
 		if (ptr) {
 			*ptr = '\0';
 			ptr++;
-		} 
+		}
 		buf = ptr;
 	}
 	*service = serv;
 	return 0;
+out_err:
+        if (fd >= 0)
+            close(fd);
+        free(buf);
+        free(filename);
+        return -1;
 }
 
 int readServices(struct service **services) {
@@ -260,7 +295,7 @@ int readServices(struct service **services) {
 	if (!(dir = opendir(RUNLEVELS "/init.d"))) {
 		fprintf(stderr, _("failed to open %s/init.d: %s\n"), RUNLEVELS,
 			strerror(errno));
-		return 1;
+		return -1;
 	}
 
 	while ((ent = readdir(dir))) {
@@ -292,7 +327,7 @@ int readServices(struct service **services) {
 }
 
 int readServiceInfo(char * name, int type, struct service * service, int honorHide) {
-    char * filename = alloca(strlen(name) + strlen(RUNLEVELS) + 50);
+    char * filename = NULL;
     int fd;
     struct service serv, serv_overrides;
     int parseret;
@@ -300,16 +335,17 @@ int readServiceInfo(char * name, int type, struct service * service, int honorHi
     if (!(type & TYPE_INIT_D))
 	goto try_xinetd;
 
-    sprintf(filename, RUNLEVELS "/init.d/%s", name);
+    asprintf(&filename, RUNLEVELS "/init.d/%s", name);
 
     if ((fd = open(filename, O_RDONLY)) < 0)
 	goto try_xinetd;
 
+    free(filename);
     parseret = parseServiceInfo(fd, name, &serv, honorHide, 0);
     if (parseret)
         return parseret;
 
-    sprintf(filename, RUNLEVELS "/chkconfig.d/%s", name);
+    asprintf(&filename, RUNLEVELS "/chkconfig.d/%s", name);
     if ((fd = open(filename, O_RDONLY)) >= 0) {
         parseret = parseServiceInfo(fd, name, &serv_overrides, honorHide, 1);
         if (parseret >= 0) {
@@ -326,18 +362,21 @@ int readServiceInfo(char * name, int type, struct service * service, int honorHi
             if (serv_overrides.isLSB || serv.isLSB) serv.isLSB = 1;
         }
     }
+    serv.currentLevels = whatLevels(name);
 
+    free(filename);
     *service = serv;
     return 0;
 
 try_xinetd:
+    free(filename);
     if (!(type & TYPE_XINETD))
 	return -1;
-    return readXinetdServiceInfo(name,service,honorHide);
+    return readXinetdServiceInfo(name,service);
 }
 
 int readServiceDifferences(char * name, int type, struct service * service, struct service * service_overrides, int honorHide) {
-    char * filename = alloca(strlen(name) + strlen(RUNLEVELS) + 50);
+    char * filename = NULL;
     int fd;
     struct service serv, serv_overrides;
     int parseret;
@@ -345,23 +384,26 @@ int readServiceDifferences(char * name, int type, struct service * service, stru
     if (!(type & TYPE_INIT_D))
 	goto try_xinetd;
 
-    sprintf(filename, RUNLEVELS "/init.d/%s", name);
+    asprintf(&filename, RUNLEVELS "/init.d/%s", name);
 
     if ((fd = open(filename, O_RDONLY)) < 0) {
 	goto try_xinetd;
     }
 
+    free(filename);
     parseret = parseServiceInfo(fd, name, &serv, honorHide, 0);
     if (parseret) {
         return parseret;
     }
 
-    sprintf(filename, RUNLEVELS "/chkconfig.d/%s", name);
+    asprintf(&filename, RUNLEVELS "/chkconfig.d/%s", name);
     if ((fd = open(filename, O_RDONLY)) >= 0) {
         parseret = parseServiceInfo(fd, name, &serv_overrides, honorHide, 1);
     } else {
-        return 1;
+        parseret = 1;
     }
+
+    free(filename);
     if (parseret) {
         return 1;
     }
@@ -371,9 +413,10 @@ int readServiceDifferences(char * name, int type, struct service * service, stru
     return 0;
 
 try_xinetd:
+    free(filename);
     if (!(type & TYPE_XINETD))
 	return -1;
-    return readXinetdServiceInfo(name,service,honorHide);
+    return readXinetdServiceInfo(name,service);
 }
 
 static struct dep *parseDeps(char *pos, char *end) {
@@ -409,19 +452,20 @@ static struct dep *parseDeps(char *pos, char *end) {
 int parseServiceInfo(int fd, char * name, struct service * service, int honorHide, int partialOk) {
     struct stat sb;
     char * bufstart, * bufstop, * start, * end, * next, *tmpbufstart;
-    struct service serv = { 
-	    	    name: NULL, 
-		    levels: -1, 
-		    kPriority: -1, 
-		    sPriority: -1, 
-		    desc: NULL, 
-		    startDeps: NULL, 
+    struct service serv = {
+		    name: NULL,
+		    levels: -1,
+		    kPriority: 100,
+		    sPriority: -1,
+		    currentLevels: 0,
+		    desc: NULL,
+		    startDeps: NULL,
 		    stopDeps: NULL,
-		    softStartDeps: NULL, 
+		    softStartDeps: NULL,
 		    softStopDeps: NULL,
 		    provides: NULL,
-		    type: TYPE_INIT_D, 
-		    isLSB: 0, 
+		    type: TYPE_INIT_D,
+		    isLSB: 0,
 		    enabled: 0
     };
     char overflow;
@@ -432,13 +476,13 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 
     bufstart = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
     if (bufstart == ((caddr_t) -1)) {
-	close(fd);	
+	close(fd);
 	return -1;
     }
 
     tmpbufstart = (char*)malloc(sb.st_size+1);
     if (tmpbufstart == NULL) {
-	close(fd);	
+	close(fd);
 	return -1;
     }
 
@@ -456,10 +500,10 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 	start = next;
 
 	while (isspace(*start) && start < bufstop) start++;
-	if (start == bufstop) break; 
+	if (start == bufstop) break;
 
 	end = strchr(start, '\n');
-	if (!end) 
+	if (!end)
 	    next = end = bufstop;
 	else
 	    next = end + 1;
@@ -469,7 +513,7 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 	start++;
 	if (!strncmp(start, "## BEGIN INIT INFO", 18))
 		    serv.isLSB = 1;
-		
+
 	while (isspace(*start) && start < end) start++;
 	if (start == end) continue;
 	if (honorHide && !strncmp(start, "hide:", 5)) {
@@ -484,7 +528,7 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 
 	if (!strncmp(start, "chkconfig:", 10)) {
 	    int spri, kpri;
-	     
+
 	    start += 10;
 	    while (isspace(*start) && start < end) start++;
 	    if (start == end) {
@@ -507,7 +551,7 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 	    }
 	    if (serv.sPriority == -1)
 			serv.sPriority = spri;
-	    if (serv.kPriority == -1)
+	    if (serv.kPriority == 100)
 			serv.kPriority = kpri;
 
 	    if (serv.levels == -1) {
@@ -532,11 +576,11 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 		}
 	} else if (!strncmp(start, "Default-Start:", 14)) {
 		char *t;
-		
+
 		start+=14;
 		while (1) {
 			int lev;
-			
+
 			lev = strtol(start, &t, 10);
 			if (t && t != start)
 				start = t;
@@ -548,11 +592,11 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 		}
 	} else if (!strncmp(start, "Default-Stop:", 13)) {
 		char *t;
-		
+
 		start+=13;
 		while (1) {
 			int lev;
-			
+
 			lev = strtol(start, &t, 10);
 			if (t && t != start)
 				start = t;
@@ -577,7 +621,7 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 	} else if (!strncmp(start, "Provides:", 9)) {
 		char *t;
 		int numdeps = 0;
-		
+
 		start+=9;
 		while (1) {
 			while (*start && isspace(*start) && start < end) start++;
@@ -599,7 +643,7 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
 			else
 				start = t;
 		}
-		
+
 	}
     }
 
@@ -613,9 +657,9 @@ int parseServiceInfo(int fd, char * name, struct service * service, int honorHid
     } else if (english_desc)
 	free (english_desc);
 
-    if (!partialOk && ((serv.levels == -1) || !serv.desc || (!serv.isLSB && (serv.sPriority == -1 || serv.kPriority == -1)))) {
+    if (!partialOk && ((serv.levels == -1) || !serv.desc || (!serv.isLSB && (serv.sPriority == -1 || serv.kPriority == 100)))) {
 	return 1;
-    } 
+    }
 
     serv.name = strdup(name);
     if (!serv.provides) {
@@ -643,7 +687,7 @@ int currentRunlevel(void) {
 
     pclose(p);
 
-    if (response[1] != ' ' || !isdigit(response[2]) || response[3] != '\n') 
+    if (response[1] != ' ' || !isdigit(response[2]) || response[3] != '\n')
 	return -1;
 
     return response[2] - '0';
@@ -717,13 +761,24 @@ int isOn(char * name, int level) {
     return 1;
 }
 
+int whatLevels(char *name) {
+    int i, ret = 0;
+
+    for (i = 0; i < 7; i++) {
+        ret |= (isOn(name, i) << i);
+    }
+    return ret;
+}
+
 int setXinetdService(struct service s, int on) {
 	int oldfd, newfd;
 	char oldfname[100], newfname[100];
 	char tmpstr[50];
 	char *buf, *ptr, *tmp;
 	struct stat sb;
-	
+        mode_t mode;
+        int r;
+
 	if (on == -1) {
 		on = s.enabled ? 1 : 0;
 	}
@@ -741,7 +796,9 @@ int setXinetdService(struct service s, int on) {
 	close(oldfd);
 	buf[sb.st_size] = '\0';
 	snprintf(newfname,100,"%s/%s.XXXXXX",XINETDDIR,s.name);
+        mode = umask(S_IRWXG | S_IRWXO);
 	newfd = mkstemp(newfname);
+        umask(mode);
 	if (newfd == -1) {
 		free(buf);
 		return -1;
@@ -752,7 +809,7 @@ int setXinetdService(struct service s, int on) {
 		if (ptr) {
 			*ptr = '\0';
 			ptr++;
-		} 
+		}
 		while (isspace(*buf)) buf++;
 		if (strncmp(buf,"disable", 7) && strlen(buf)) {
 			write(newfd,tmp,strlen(tmp));
@@ -766,9 +823,12 @@ int setXinetdService(struct service s, int on) {
 		buf = ptr;
 	}
 	close(newfd);
-	chmod(newfname,0644);
 	unlink(oldfname);
-	return(rename(newfname,oldfname));
+        r = rename(newfname,oldfname);
+        if (selinux_restore(oldfname) != 0)
+                fprintf(stderr, _("Unable to set selinux context for %s: %s\n"), oldfname,
+		strerror(errno));
+	return(r);
 }
 
 int doSetService(struct service s, int level, int on) {
@@ -795,6 +855,414 @@ int doSetService(struct service s, int level, int on) {
 	return 1;
     }
 
-    return 0; 
+    return 0;
 }
 
+int systemdIsInit() {
+    char *path = realpath("/sbin/init", NULL);
+    char *base = NULL;
+
+    if (!path)
+        return 0;
+    base = basename(path);
+    if (!base)
+        return 0;
+    if (strcmp(base,"systemd"))
+        return 0;
+    return 1;
+}
+
+int systemdActive() {
+    struct stat a, b;
+
+    if (lstat("/sys/fs/cgroup", &a) < 0)
+        return 0;
+    if (lstat("/sys/fs/cgroup/systemd", &b) < 0)
+        return 0;
+    if (a.st_dev == b.st_dev)
+        return 0;
+    if (!systemdIsInit())
+        return 0;
+    return 1;
+}
+
+
+int isOverriddenBySystemd(const char *service) {
+    char *p;
+    int rc = 0;
+
+    asprintf(&p, SYSTEMD_SERVICE_PATH "/%s.service", service);
+
+    if (access(p, F_OK) >= 0) {
+        rc = 1;
+        goto out;
+    }
+    free(p);
+    asprintf(&p, SYSTEMD_LOCAL_SERVICE_PATH "/%s.service", service);
+    if (access(p, F_OK) >= 0) {
+        rc = 1;
+    }
+out:
+    free(p);
+    return rc;
+}
+
+int isSocketActivatedBySystemd(const char *service) {
+    char *p;
+    char *s;
+    int rc = 0;
+
+    asprintf(&p, SYSTEMD_SERVICE_PATH "/%s@.service", service);
+    asprintf(&s, SYSTEMD_SERVICE_PATH "/%s.socket", service);
+
+    if (access(p, F_OK) >= 0 && access(s, F_OK) >= 0) {
+        rc = 1;
+        goto out;
+    }
+    free(p);
+    free(s);
+
+    asprintf(&p, SYSTEMD_LOCAL_SERVICE_PATH "/%s@.service", service);
+    asprintf(&s, SYSTEMD_LOCAL_SERVICE_PATH "/%s.socket", service);
+    if (access(p, F_OK) >= 0 && access(s, F_OK) >= 0) {
+        rc = 1;
+    }
+out:
+    free(p);
+    free(s);
+    return rc;
+}
+
+int isEnabledInSystemd(const char *service) {
+        char *c = NULL;
+        int r;
+        if(!systemdActive())
+                return 0;
+
+        if(asprintf(&c, "systemctl is-enabled %s.service --quiet", service)<0)
+                return 0;
+
+        r = system(c);
+        free(c);
+
+        return r == 0 ? 1 : 0;
+}
+
+int readSystemdUnitProperty(char *name, char *property, char **value) {
+        FILE *sys = NULL;
+        char *cmd = NULL;
+        char *line = NULL;
+        char *t = NULL;
+        int i;
+        int r = 0;
+        int c;
+
+        if (value == NULL)
+                return -1;
+
+        r = asprintf(&cmd, "systemctl show %s -p %s", name, property);
+        if (r < 0)
+                goto finish;
+        else
+                r = 0;
+
+        sys = popen(cmd, "r");
+        if (!sys) {
+                r = -1;
+                goto finish;
+        }
+
+        for (i = 0; (c = fgetc(sys)) != EOF; i++) {
+                t = realloc(line, i + 2);
+                if (t == NULL) {
+                        r = -ENOMEM;
+                        goto finish;
+                }
+                line = t;
+                line[i] = (char) c;
+                line[i + 1] = 0;
+        }
+
+        if (line == NULL) {
+                r = -ENOENT;
+                goto finish;
+        }
+        t = strchr(line, '\n');
+        if (t)
+                *t = 0;
+
+        t = strchr(line, '=');
+        if (t == NULL) {
+                r = -1;
+                goto finish;
+        }
+
+        *value = strdup(t + 1);
+        if (value == NULL)
+                r = -ENOMEM;
+
+finish:
+
+
+        if (sys)
+                pclose(sys);
+        if (cmd)
+                free(cmd);
+        if (line)
+                free(line);
+
+        return r;
+}
+
+
+static int cmpstringp(const void *p1, const void *p2){
+        return strcmp(* (char * const *) p1, * (char * const *) p2);
+}
+
+
+
+int runlevelsToTargets (int runlevels, char ***targets, int *n_targets) {
+        char runlevel[] = "runlevelX.target";
+        char **ret = NULL;
+        char **tmp;
+        int n_ret = 0;
+        int found = 0;
+        char *t;
+        int i,j;
+        int r;
+
+        for (i=0; i<=6; i++) {
+                if (1<<i & runlevels){
+                      runlevel[8] = '0' + i;
+                      r = readSystemdUnitProperty(runlevel, "Id", &t);
+                      if (r < 0)
+                              goto fail;
+                      for (j=0; j<n_ret; j++) {
+                              if(!strcmp(t, ret[j])) {
+                                      found = 1;
+                                      break;
+                              }
+                      }
+
+                      if (!found) {
+                              tmp = (char **) realloc(ret, sizeof(char *)*(n_ret+1));
+                              if (tmp == NULL) {
+                                      r = -ENOMEM;
+                                      goto fail;
+                              }
+                              ret = tmp;
+                              ret[n_ret] = t;
+                              n_ret++;
+                      } else
+                              free(t);
+
+                      found = 0;
+                }
+        }
+
+
+        *targets = ret;
+        *n_targets = n_ret;
+
+        return 0;
+
+fail:
+        if(ret){
+                for (j = 0; j < n_ret; j++)
+                        free(ret[j]);
+                free(ret);
+        }
+        return r;
+
+}
+
+int serviceNameToUnit(const char *service, char **unit) {
+//translation $* -> *.target taken from systemd src/core/special.h
+        char *ret = NULL;
+        if (!strcmp(service, "$local_fs")) {
+                ret = strdup("local-fs.target");
+        } else if (!strcmp(service, "$network")){
+                ret = strdup("network.target");
+        } else if (!strcmp(service, "$named")){
+                ret = strdup("nss-lookup.target");
+        } else if (!strcmp(service, "$portmap")){
+                ret = strdup("rpcbind.target");
+        } else if (!strcmp(service, "$remote_fs")){
+                ret = strdup("remote-fs.target");
+        } else if (!strcmp(service, "$syslog")){
+                //not nice
+                ret = strdup("systemd-journald.service");
+        } else if (!strcmp(service, "$time")){
+                ret = strdup("time-sync.target");
+        } else {
+                asprintf(&ret, "%s.service", service);
+        }
+
+        if (ret == NULL)
+                return -ENOMEM;
+
+        *unit = ret;
+        return 0;
+}
+
+int unitGetReverseDeps(char *unit, char ***deps, int *n_deps) {
+        FILE *sys = NULL;
+        char *cmd = NULL;
+        char **ret = NULL;
+        int n_ret = 0;
+        char *t = NULL;
+        char **tt = NULL;
+        int i,j;
+        int r = 0;
+        int c = 0;
+
+        r = asprintf(&cmd, "systemctl list-dependencies --reverse --all --plain %s", unit);
+        if (r < 0)
+                goto finish;
+        else
+                r = 0;
+
+        sys = popen(cmd, "r");
+
+        if (!sys) {
+                r = -1;
+                goto finish;
+        }
+
+        for (i = 0; !feof(sys); i++) {
+                tt = realloc(ret, sizeof(char *) * (n_ret + 1));
+                if (tt == NULL) {
+                        r = -ENOMEM;
+                        goto finish;
+                }
+                ret = tt;
+                ret[n_ret] = NULL;
+
+                for (j = 0; (c = fgetc(sys)) != EOF && c != '\n'; j++) {
+                        t = realloc(ret[n_ret], j + 2);
+                        if (t == NULL) {
+                                if(ret[n_ret])
+                                        n_ret++;
+                                r = -ENOMEM;
+                                goto finish;
+                        }
+                        ret[n_ret] = t;
+                        ret[n_ret][j] = (char) c;
+                        ret[n_ret][j + 1] = 0;
+                }
+
+                if(ret[n_ret])
+                        n_ret++;
+        }
+
+
+        *deps = ret;
+        *n_deps = n_ret;
+
+
+finish:
+        if (sys)
+                pclose(sys);
+        if (cmd)
+                free(cmd);
+
+        if(r) {
+                if(ret) {
+                        for (i = 0; i < n_ret; i++)
+                                free(ret[i]);
+                        free(ret);
+                }
+        }
+        return r;
+}
+
+int systemdIsSimilarlyConfiguredDeps(char **deps1, int n_deps1, char **deps2, int n_deps2) {
+        int i,j;
+        int r;
+        if (n_deps1 > n_deps2)
+                return 0;
+
+        for (i=0, j=0; i < n_deps1; i++){
+                while ((r = strcmp(deps1[i], deps2[j])) > 0)
+                        if(++j >= n_deps2)
+                                return 0;
+                if(r < 0)
+                        return 0;
+        }
+        return 1;
+}
+
+int unitIsSimilarlyConfigure(char **star, int n_star, char *unit) {
+        char **deps=NULL;
+        int n_deps=0;
+        int r;
+        int i;
+
+        r = unitGetReverseDeps(unit, &deps, &n_deps);
+
+        if (r)
+                return 0;
+
+        qsort(deps, n_deps, sizeof(char *), cmpstringp);
+
+        r = systemdIsSimilarlyConfiguredDeps(star, n_star, deps, n_deps);
+
+        if(deps) {
+                for (i = 0; i < n_deps; i++)
+                        free(deps[i]);
+                free(deps);
+        }
+
+        return r;
+
+
+}
+
+void checkSystemdDependencies(struct service *s) {
+        char **star = NULL;
+        int n_star = 0;
+        char *unit;
+        int r = 0;
+        int i;
+
+
+        r = runlevelsToTargets (s->levels , &star, &n_star);
+
+        if(r)
+                return;
+
+        qsort(star, n_star, sizeof(char *), cmpstringp);
+
+
+        if (s->startDeps) {
+                for (i = 0; s->startDeps[i].name; i++) {
+                        if (!s->startDeps[i].handled) {
+                                r = serviceNameToUnit(s->startDeps[i].name, &unit);
+                                if(r)
+                                        goto finish;
+                                s->startDeps[i].handled = unitIsSimilarlyConfigure(star, n_star, unit);
+                                free(unit);
+                        }
+                }
+        }
+        if (s->stopDeps) {
+                for (i = 0; s->stopDeps[i].name; i++) {
+                        if (!s->stopDeps[i].handled) {
+                                r = serviceNameToUnit(s->stopDeps[i].name, &unit);
+                                if(r)
+                                        goto finish;
+                                s->stopDeps[i].handled = unitIsSimilarlyConfigure(star, n_star, unit);
+                                free(unit);
+                        }
+                }
+        }
+
+
+finish:
+
+        if(star) {
+                for (i = 0; i < n_star; i++)
+                        free(star[i]);
+                free(star);
+        }
+}
